@@ -83,12 +83,17 @@ def setup_arg_parsing():
     parser_getresources.add_argument('--manage', action='store_true', help='Manage the matching resources')
     parser_getresources.add_argument('--filter', required=False, help='Post-query filter on resources.  Python expression that will evaluate to True or False such as alert["resource"]["name"].startswith("prod")')
 
-    parser_importresources = subparsers.add_parser('importresources', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Import resources from a spreadsheet.  Custom attr field names should be prefixed with .tag and the custom attribute names should be created in advance via the UI.  Any resource with an Action Processing column value of Delete will deleted.")
+    parser_importresources = subparsers.add_parser('importresources', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Import resources from a spreadsheet.  Custom attr field names should be prefixed with \"tag.\" and the custom attribute names should be created in advance via the UI.  Any resource with an Action Processing column value of Delete will be deleted.")
     parser_importresources.add_argument('--commit', action='store_true', help='Make the actual updates on the platform.  If not specified, only error checking and import simulation will occur.')
     parser_importresources.add_argument('--addvalues', action='store_true', help='If new values are found in the spreadsheet, add the value definitions on the fly.  Otherwise new values will be considered an error.')
     parser_importresources.add_argument('--writeblanks', action='store_true', help='When no value is provided in the spreadsheet for a resource, remove any existing value for that resource on the platform.  If not specified then no action is taken for empty values.')
     parser_importresources.add_argument('--filename', required=False, help='Name of excel file to import (.xlsx extension will be added if not specified.)')
+    parser_importresources.add_argument('--nocreate', action='store_true', help='Don\'t create new resources.')
 
+    parser_addpropasmetriclabel = subparsers.add_parser('addpropasmetriclabel', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Add a resource property as a custom attribute value of the matched resources, with metric label option selected")
+    parser_addpropasmetriclabel.add_argument('--query', required=False, help='Query String to filter resources as per https://develop.opsramp.com/resource-management/tenants-tenantid-resources-search')
+    parser_addpropasmetriclabel.add_argument('--property', required=False, help='Resource property in dot notation to be with values to be added as a custattr values, for example attributes.accountName (Must be in content returned by getresources)')
+    parser_addpropasmetriclabel.add_argument('--custattrname', required=True, help='Custom attribute name - must already be existing (i.e. defined via UI.).  This command only adds new values to it if needed.')
 
 
     # Service Maps
@@ -130,6 +135,8 @@ def setup_arg_parsing():
     parser_exportcustattrfile.add_argument('--search', required=False, help='Search String to filter resources as it would be entered under Resources -> Search')
     parser_exportcustattrfile.add_argument('--filter', required=False, help='Post-query filter on resources.  Python expression that will evaluate to True or False such as alert["resource"]["name"].startswith("prod")')
     parser_exportcustattrfile.add_argument('--filename', required=False, help='Name of excel file to generate (.xlsx extension will be added)')
+    parser_exportcustattrfile.add_argument('--attrs', required=False, help='Comma separated list of Custom Attribute names to include (no custom attrs will be exported if not specified)')
+    parser_exportcustattrfile.add_argument('--props', required=False, help='Comma separated list of Resource Properties to include (Use dot-walk notation based on getresources json payload)')
 
     parser_importcustattrfile = subparsers.add_parser('importcustattrfile', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Import an Excel file containing custom attribute values")
     parser_importcustattrfile.add_argument('--commit', action='store_true', help='Make the actual updates on the platform.  If not specified, only error checking and import simulation will occur.')
@@ -139,6 +146,14 @@ def setup_arg_parsing():
 
     parser_getcustomattrs = subparsers.add_parser('getcustomattrs', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Get custom attribute definitions")
 
+    parser_importfromdatasource= subparsers.add_parser('importfromdatasource', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Import Resources & Custom Attributes from a data source.")
+    parser_importfromdatasource.add_argument('--job', required=True, help='Name of yaml job definition file')
+    parser_importfromdatasource.add_argument('--commit', action='store_true', help='Make the actual updates on the platform.  If not specified, only error checking and import simulation will occur.')
+    parser_importfromdatasource.add_argument('--addvalues', action='store_true', help='If new values are found in the data source, add the value definitions on the fly.  Otherwise new values will be considered an error.')
+    parser_importfromdatasource.add_argument('--writeblanks', action='store_true', help='When no value is provided in the data source record for a resource, remove any existing value for that resource on the platform.  If not specified then no action is taken for empty values.')
+    parser_importfromdatasource.add_argument('--nocreate', action='store_true', help='Don\'t create new resources.')
+    parser_importfromdatasource.add_argument('--metricLabel', action='store_true', help='Add new attribute values with metricLabel=true (i.e. they will be used as metric labels).')
+
     # Discovery Profiles
     parser_getdiscoprofile = subparsers.add_parser('getdiscoprofile', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Get discovery profile definition")
     parser_getdiscoprofile.add_argument('--id', required=True, help='Discovery profile ID')
@@ -161,14 +176,20 @@ def setup_arg_parsing():
     parser_migratealertesc.add_argument('--update', action='store_true', help='Used for bulk updates, will only work if --env and --to_env are the same.  Try to update existing policies instead of creating new ones.')
     parser_migratealertesc.add_argument('--setactive', required=False, help='Specify ON or OFF.  Will force all policies created on the target to be ON or OFF.  Otherwise will be set the same as the source.')
 
-    # Monitoring
+    # Device Management Policies get_device_management_policies
+    parser_getdevmgtpol = subparsers.add_parser('getdevmgtpol', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Search and get Escalation Policy definitions")
+    parser_getdevmgtpol.add_argument('--query', required=False, help='Query String to filter alerts')
+    parser_getdevmgtpol.add_argument('--details', action='store_true', help='Get the full details for all matched policies')
+    parser_getdevmgtpol.add_argument('--count', action='store_true', help='Only show the count of matching alerts')
+    parser_getdevmgtpol.add_argument('--filter', required=False, help='Post-query filter on alerts.  Python expression that will evaluate to True or False such as alert["resource"]["name"].startswith("prod")')
+
+
+    # Templates
 
     parser_gettemplates = subparsers.add_parser('gettemplates', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Clone monitoring templates")
     parser_gettemplates.add_argument('--query', help="Query search criteria")
     parser_gettemplates.add_argument('--brief', action='store_true', help='Output only the template names not the full definitions')
 
-
-
     parser_clonetemplates = subparsers.add_parser('clonetemplates', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Clone monitoring templates")
     parser_clonetemplates_src = parser_clonetemplates.add_mutually_exclusive_group(required=True)
     parser_clonetemplates_src.add_argument('--name', help="Name of the Monitoring Template to clone")
@@ -177,6 +198,24 @@ def setup_arg_parsing():
     parser_clonetemplates_target.add_argument('--prefix', help="New template name will be same as original name but with this text prepended")
     parser_clonetemplates_target.add_argument('--copyname', help="New template will have this as its name")
 
+    # Monitoring and Metrics
+
+    parser_metricsql = subparsers.add_parser('metricsql', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Run a MetricsQL/PromQL query")
+    parser_metricsql.add_argument('--instant_query', help="MetricsQL/PromQL Query")
+
+    parser_setbaselinethreshold = subparsers.add_parser('setbaselinethreshold', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Set a new override threshold for a resource/component that matches a given MetricQL query")
+    parser_setbaselinethreshold.add_argument('--template', required=True, help="Name of template that metric is assigned from")
+    parser_setbaselinethreshold.add_argument('--metric', required=True, help="Technical metric name as listed in the template definition") 
+    parser_setbaselinethreshold.add_argument('--instant_query', required=True, help="PromQL instant query returning a vector result (i.e. must aggregate over time)")
+    parser_setbaselinethreshold.add_argument('--warnop', choices=('NONE', 'LESS_THAN', 'LESS_THAN_EQUAL', 'EQUAL', 'GREATER_THAN_EQUAL', 'GREATER_THAN'), help="Comparison operator for Warning")
+    parser_setbaselinethreshold.add_argument('--critop', choices=('NONE', 'LESS_THAN', 'LESS_THAN_EQUAL', 'EQUAL', 'GREATER_THAN_EQUAL', 'GREATER_THAN'), help="Comparison operator for Warning")
+    parser_setbaselinethreshold.add_argument('--warnthresh', type=float, help='Warning threshold value')
+    parser_setbaselinethreshold.add_argument('--critthresh', type=float, help='Critical threshold value')
+    parser_setbaselinethreshold.add_argument('--warnrepeat', type=int, default=1, help='Warning repeat count')
+    parser_setbaselinethreshold.add_argument('--critrepeat', type=int, default=1, help='Critical repeat count')
+    parser_setbaselinethreshold.add_argument('--raisealert', required=True, choices=('on', 'off'), help='Whether to raise an alert or not')
+    parser_setbaselinethreshold.add_argument('--availability', action='store_true', help='Make the metric an availability metric')
+
     # Integrations
     parser_getintegrations = subparsers.add_parser('getintegrations', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Get installed Integration definitions")
     parser_getintegrations.add_argument('--query', required=False, help='Query String to filter integrations')
@@ -201,12 +240,17 @@ def setup_arg_parsing():
     parser_addazurearmintegration.add_argument('--azclient', required=True, help='Azure Client ID')
     parser_addazurearmintegration.add_argument('--azsecret', required=True, help='Azure Secret Key')
 
-    parser_addazureadmintegration = subparsers.add_parser('addazureasmintegration', parents=[env_parser,discp_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Install a new Azure ADM integration")
-    parser_addazureadmintegration.add_argument('--azname', required=True, help='displayName for the integration')
-    parser_addazureadmintegration.add_argument('--azsub', required=True, help='Azure Subscription ID')
-    parser_addazureadmintegration.add_argument('--azcert', required=True, help='Azure Management Certificate')
-    parser_addazureadmintegration.add_argument('--azkspw', required=True, help='Azure Keystore Password')
+    parser_addazureasmintegration = subparsers.add_parser('addazureasmintegration', parents=[env_parser,discp_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Install a new Azure ASM integration")
+    parser_addazureasmintegration.add_argument('--azname', required=True, help='displayName for the integration')
+    parser_addazureasmintegration.add_argument('--azsub', required=True, help='Azure Subscription ID')
+    parser_addazureasmintegration.add_argument('--azcert', required=True, help='Azure Management Certificate')
+    parser_addazureasmintegration.add_argument('--azkspw', required=True, help='Azure Keystore Password')
 
+    # Kubernetes Integrations
+    parser_addkubernetesintegration = subparsers.add_parser('addkubernetesintegration', parents=[env_parser], formatter_class=argparse.ArgumentDefaultsHelpFormatter, help="Install a new Kubernetes integration")
+    parser_addkubernetesintegration.add_argument('--name', required=True, help='Name for the integration, or if prefixed with @ a text file with a list of names.')
+    parser_addkubernetesintegration.add_argument('--type', choices=('cloud','onPrem'), required=True, help='Kubernetes deployment type')
+    parser_addkubernetesintegration.add_argument('--engine', required=True, choices=('Docker','ContainerD','CRI-O'), help='Container engine type')
 
     return parser
 
