@@ -2,8 +2,10 @@ from genericpath import exists
 import yaml
 import sys
 import requests
+from requests.exceptions import HTTPError
 import json
 import time
+from http import HTTPStatus
 from datetime import datetime
 #import opsramp.binding as po_binding
 #import opsramp.tenant as po_tenant
@@ -11,6 +13,8 @@ from datetime import datetime
 
 class OpsRampEnv:
 
+    MAX_NEW_CUSTOM_ATTR_VALUES = 50
+
     OPS_ALERT_SEARCH_ATTRIBUTES = [
         'states',
         'startDate',
@@ -41,11 +45,37 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
-        response = requests.request("POST", url, headers=headers, data=json.dumps(obj), verify=self.isSecure)
-        if int(response.headers['x-ratelimit-remaining']) < 2 :
-            sleeptime = int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3
+        retries = 3
+        retry_codes = [
+            HTTPStatus.TOO_MANY_REQUESTS,
+            HTTPStatus.INTERNAL_SERVER_ERROR,
+            HTTPStatus.BAD_GATEWAY,
+            HTTPStatus.SERVICE_UNAVAILABLE,
+            HTTPStatus.GATEWAY_TIMEOUT,
+        ]
+        response:requests.Response
+        for n in range(retries):
+            try:
+                response = requests.request("POST", url, headers=headers, data=json.dumps(obj), verify=self.isSecure)
+                response.raise_for_status()
+                break
+            except HTTPError as exc:
+                code = exc.response.status_code                
+                if code in retry_codes:
+                    # retry after n seconds
+                    time.sleep(n)
+                    continue
+                raise
+        
+
+        if 'x-ratelimit-remaining' in response.headers and int(response.headers['x-ratelimit-remaining']) < 2 :
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
+
             print(f'Sleeping for {str(sleeptime)} sec to wait for API throttling limit to reset..')
             time.sleep(sleeptime)
         try:    
@@ -53,8 +83,113 @@ class OpsRampEnv:
         except Exception as e:
             return {'success': response.ok, 'text': response.text}
 
+    def put_object(self, path, obj):
+        token = self.get_token()
+        url = self.env['url'] + path
+        headers = {
+            'Content-Type': 'application/json',
+            'Accept': 'application/json',
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
+        }
+        response = requests.request("PUT", url, headers=headers, data=json.dumps(obj), verify=self.isSecure)
+        if 'x-ratelimit-remaining' in response.headers and int(response.headers['x-ratelimit-remaining']) < 2 :
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
+            print(f'Sleeping for {str(sleeptime)} sec to wait for API throttling limit to reset..')
+            time.sleep(sleeptime)
+        try:    
+            return response.json()
+        except Exception as e:
+            return {'success': response.ok, 'text': response.text}
+        
+    def get_with_pagination(self, path:str, result_array_key:str='results', params:dict={}, page_no:int=1, page_size:int=500):
+        token = self.get_token()
+        url = self.env['url'] + path
+        headers = {
+            'Content-Type': 'application/json',
+            'Accept': 'application/json',
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
+        }
+
+        params['pageSize'] =  page_size
+        params['pageNo'] = page_no
+
+        
+        response = requests.request("GET", url, headers=headers, params=params, verify=self.isSecure)
+        if 'x-ratelimit-remaining' in response.headers and int(response.headers['x-ratelimit-remaining']) < 2 :
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
+            print(f'Sleeping for {str(sleeptime)} sec to wait for API throttling limit to reset..')
+            time.sleep(sleeptime)
+        bodydict = response.json()
+        results = []
+        if result_array_key in bodydict:
+            results = bodydict[result_array_key]
+        if 'nextPage' in bodydict and bodydict['nextPage']:
+            return results + self.get_with_pagination(path=path, result_array_key=result_array_key, params=params, page_no=page_no+1, page_size=page_size)
+        else:
+            return results
+        
+    def do_opsql_query(self, object_type:str, fields:list=None, filter_criteria:str=None, aggregate:str=None, group_by:list=None, sort_by:str=None, page_no:int=1, page_size:int=1000):
+        url = f'/opsql/api/v7/tenants/{self.env["tenant"]}/queries'
+        payload = {}
+        payload['objectType'] = object_type
+        if fields:
+            payload['fields'] = fields
+        if filter_criteria: 
+            payload['filterCriteria'] = filter_criteria
+        if aggregate:
+            payload['aggregateFunction'] = aggregate
+        if group_by:
+            payload['groupBy'] = group_by
+        if sort_by:
+            payload['sortBy'] = sort_by
+        if page_no:
+            payload['pageNo'] = page_no
+        if page_size:
+            payload['pageSize'] = page_size
+
+        bodydict = self.post_object(url, payload)
+        results = []
+        if 'results' in bodydict:
+            results = bodydict['results']
+        if 'nextPage' in bodydict and bodydict['nextPage']:
+            results + self.do_opsql_query(object_type, fields, filter_criteria, aggregate, group_by, sort_by, page_no+1, page_size)
+        return results
+            
+
+
+
+    def do_instant_metric_query(self, query):
+        url = f'/metricsql/api/v7/tenants/{self.env["tenant"]}/metrics/latest'
+        payload = {
+            "query": query 
+        }
+        metrics_instant_query_response = self.post_object(url, payload)
+        if 'status' in metrics_instant_query_response and metrics_instant_query_response['status'] == 'success':
+            return metrics_instant_query_response['data']['result']
+        else:
+            raise Exception(json.stringify(metrics_instant_query_response))
+        
+    def do_resource_threshold_assign_multi(self, resource_id, object):
+        url = f'/api/v3/tenants/{self.env["tenant"]}/resource/{resource_id}/thresholds'
+        return self.put_object(url, object)
+    
+    def get_device_management_policies(self):
+        return self.get_objects("deviceManagementPolicies")
+
+
     def get_objects(self, obtype, page=1, queryString=None, searchQuery=None,countonly=False, itemId=None, filter=None):
 
+        # Moved to v3 API
+        if obtype == "customAttributes":
+            return self.get_tags(filter_criteria=searchQuery)
+
+
         endpoints = {
             "clients": self.env['partner'] + "/clients/" + self.env['tenant'],
             "incidentCustomFields": self.env['tenant'] + "/customFields/INCIDENT",
@@ -71,7 +206,8 @@ class OpsRampEnv:
             "templates": self.env['tenant'] + "/monitoring/templates/search",
             "integration": self.env['tenant'] + "/integrations/installed/" + str(itemId),
             "integrations": self.env['tenant'] + "/integrations/installed/search",
-            "incident": f'{self.env["tenant"]}/incidents/{itemId}'
+            "incident": f'{self.env["tenant"]}/incidents/{itemId}',
+            "deviceManagementPolicies": f'{self.env["tenant"]}/policies/management'
         }
 
         url = self.env['url'] + "/api/v2/tenants/" + endpoints[obtype]
@@ -80,7 +216,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -102,8 +239,10 @@ class OpsRampEnv:
             params['pageSize'] = 100
 
         response = requests.request("GET", url, headers=headers, verify=self.isSecure, params=params)
-        if int(response.headers['x-ratelimit-remaining']) < 2 :
-            sleeptime = int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3
+        if 'x-ratelimit-remaining' in response.headers and int(response.headers['x-ratelimit-remaining']) < 2 :
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
             print(f'Sleeping for {str(sleeptime)} sec to wait for API throttling limit to reset..')
             time.sleep(sleeptime)
         try:
@@ -146,7 +285,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         #removeattrs = ['id', 'uniqueId', 'name', 'createdDate', 'updatedDate', ]
@@ -186,7 +326,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
         print("Sending: %s" % (json.dumps(alert)))
         response = requests.request("POST", url, headers=headers, data=json.dumps(alert), verify=self.isSecure)
@@ -232,7 +373,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
         params = {
         }
@@ -248,7 +390,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         data = {
@@ -274,7 +417,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -303,7 +447,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -371,7 +516,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
         print("Updating %s: %s" % (id, json.dumps(update)))
         response = requests.request("POST", url, headers=headers, data=json.dumps(update), verify=self.isSecure)
@@ -383,7 +529,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -404,7 +551,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -465,7 +613,8 @@ class OpsRampEnv:
         }
         headers = {
             'Content-Type': 'application/x-www-form-urlencoded',
-            'Accept': 'application/json'
+            'Accept': 'application/json',
+            'user-agent': 'opsrampcli'
         }
 
         try:
@@ -474,7 +623,7 @@ class OpsRampEnv:
             #print(response.text)
             return response.json()['access_token']
         except requests.exceptions.ConnectionError as err:
-            print(f'\nUnable to connect to {self.env["url"]}')
+            print(f'\nUnable to connect to {self.env["url"]}: {response}')
             print(f'Please check that this is the correct url and is resolvable/reachable!\n')
             sys.exit(1)
         except Exception as err:
@@ -488,7 +637,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("GET", url, headers=headers, verify=self.isSecure)
@@ -501,7 +651,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         params = {
@@ -540,7 +691,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("GET", url, headers=headers, verify=self.isSecure, params=params)
@@ -559,7 +711,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("POST", url, headers=headers, data=json.dumps(policy), verify=self.isSecure)
@@ -571,34 +724,58 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("POST", url, headers=headers, data=json.dumps(policy), verify=self.isSecure)
         return response.json()
+
+    def get_tags(self, filter_criteria:str=None):
+        return self.do_opsql_query(object_type='tag', fields=['name','id'], filter_criteria=filter_criteria)
     
-    def add_custom_attr_value(self, attributeId, values, description=""):
+    def get_tag_values(self, tag_id):
+        path = f'/api/v3/tenants/{self.env["tenant"]}/tags/{tag_id}/values'
+        return self.get_with_pagination(path)
+    
+    def update_tag_value(self, tag_id:str, value_id:str, description:str="", is_metric_label:bool=False ):
+        path = f'/api/v3/tenants/{self.env["tenant"]}/tags/{tag_id}/values/{value_id}'
+        obj = {
+            "description": description,
+            "metricLabel": is_metric_label            
+        }
+        resp = self.put_object(path, obj)
+        return resp
+
+
+    def add_custom_attr_value(self, attributeId, values, description="", is_metric_label:bool=False):
         if type(values) == str:
             values = [values]
+
+        if len(values) > self.MAX_NEW_CUSTOM_ATTR_VALUES:
+            results = []
+            chunks = [values[i:i + self.MAX_NEW_CUSTOM_ATTR_VALUES] for i in range(0, len(values), self.MAX_NEW_CUSTOM_ATTR_VALUES)]
+            for chunk in chunks:
+                results += self.add_custom_attr_value(attributeId, chunk, description, is_metric_label)
+            return results
         if type(description) == str:
             description = [description]
         if len(values) > len(description):
             for i in range(len(description), len(values)):
                 description.append("")
 
-        path = f'/api/v2/tenants/{self.env["tenant"]}/customAttributes/{attributeId}'
-        obj = {}
+        path = f'/api/v3/tenants/{self.env["tenant"]}/tags/{attributeId}/values'
         customAttributeValues = []
         for i,val in enumerate(values):
             customAttributeValues.append({
                     "value": val,
-                    "description": description[i]
+                    "description": description[i],
+                    "metricLabel": is_metric_label
             })
-        obj['customAttributeValues'] = customAttributeValues
-        return self.post_object(path, obj)
+        return self.post_object(path, customAttributeValues)
 
     def unset_custom_attr_on_devices(self, attr_id, value_id, device_ids):
-        path = f'/api/v2/tenants/{self.env["tenant"]}/customAttributes/{attr_id}/values/{value_id}/devices'
+        path = f'/api/v3/tenants/{self.env["tenant"]}/tags/{attr_id}/untagged-entities'
         devices = []
         if isinstance(device_ids, list):
             devices = device_ids
@@ -609,14 +786,17 @@ class OpsRampEnv:
 
         payload = []
         for device in devices:
-            payload.append({"id": device})
+            payload.append({
+                "entityType": "resource",
+                "entityId": device})
 
         token = self.get_token()
         url = self.env['url'] + path
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("DELETE", url, headers=headers, data=json.dumps(payload), verify=self.isSecure)
@@ -633,9 +813,12 @@ class OpsRampEnv:
 
         payload = []
         for device in devices:
-            payload.append({"id": device})
+            payload.append({
+                "entityType": "resource",
+                "entityId": device
+                })
 
-        path = "/api/v2/tenants/" + self.env['tenant'] + "/customAttributes/" + str(attr_id) + "/values/" + str(value_id) + "/devices"
+        path = "/api/v3/tenants/" + self.env['tenant'] + "/tags/" + str(attr_id) + "/values/" + str(value_id) + "/tagged-entities"
         return self.post_object(path, payload)
    
 
@@ -657,7 +840,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("DELETE", url, headers=headers, data=json.dumps(payload), verify=self.isSecure)
@@ -671,7 +855,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("POST", url, headers=headers, verify=self.isSecure)
@@ -681,7 +866,9 @@ class OpsRampEnv:
 
 
         if int(response.headers['x-ratelimit-remaining']) < 2 :
-            sleeptime = int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
             print(f'Sleeping for {str(sleeptime)} sec..')
             time.sleep(sleeptime)
         return responseobj
@@ -692,7 +879,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("POST", url, headers=headers, verify=self.isSecure, data=json.dumps(resource_dict))
@@ -700,15 +888,28 @@ class OpsRampEnv:
 
 
         if int(response.headers['x-ratelimit-remaining']) < 2 :
-            sleeptime = int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
             print(f'Sleeping for {str(sleeptime)} sec..')
             time.sleep(sleeptime)
         return responseobj
 
     def update_resource(self, resourceId:str, resource_dict:dict):
         token = self.get_token()
+        resourceName:str = resource_dict["resourceName"]
+        try:
+            del(resource_dict["resourceName"])
+        except:
+            pass
+        try:
+            del(resource_dict["tags"])
+        except:
+            pass
         path = f'/api/v2/tenants/{self.env["tenant"]}/resources/{resourceId}'
-        return self.post_object(path, resource_dict)
+        retval = self.post_object(path, resource_dict)
+        resource_dict["resourceName"] = resourceName
+        return retval
 
     def delete_resource(self, resourceId):
         token = self.get_token()
@@ -716,7 +917,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
 
         response = requests.request("DELETE", url, headers=headers, verify=self.isSecure)
@@ -726,7 +928,9 @@ class OpsRampEnv:
 
 
         if int(response.headers['x-ratelimit-remaining']) < 2 :
-            sleeptime = int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3
+            sleeptime = abs(int(response.headers['x-ratelimit-reset']) - int(datetime.now().timestamp()) + 3)
+            if sleeptime > 60:
+                sleeptime = sleeptime % 60
             print(f'Sleeping for {str(sleeptime)} sec..')
             time.sleep(sleeptime)
         return responseobj
@@ -740,7 +944,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
         svcgrouparr = [svcgroup]
         response = requests.request("POST", url, headers=headers, verify=self.isSecure, data=json.dumps(svcgrouparr))
@@ -757,7 +962,8 @@ class OpsRampEnv:
         headers = {
             'Content-Type': 'application/json',
             'Accept': 'application/json',
-            'Authorization': 'Bearer ' + token
+            'Authorization': 'Bearer ' + token,
+            'user-agent': 'opsrampcli'
         }
         link = [
             {
